//**********************
// getIR.c for Raspberry Pi
//
// (1)NEC format
// IR carrier=38KHz
// Time unit=0.56ms
// logical 0 = on 1T + off 1T
// logical 1 = on 1T + off 3T
// reader=on 16T + off 8T
// stop=on 1T
// frame=108ms
//
// (2)AEHA format
// IR carrier=33 - 40KHz
// Time unit=0.35 - 0.50ms
// logical 0 = on 1T + off 1T
// logical 1 = on 1T + off 3T
// reader=on 8T + off 4T
// trailer=on 1T + 8ms
//
// (3)SONY format
// IR carrier=40KHz
// Time unit=0.6ms
// logical 0 = off 1T + on 1T
// logical 1 = off 1T + on 2T
// reader=on 4T
// frame=45ms
//
// caution:
// no detecting repeat code, return bits=0;
//
// (C)Copyright 2015-2024 All rights reserved by Y.Onodera
// http://einstlab.web.fc2.com
//**********************

#include <stdio.h>
#include <unistd.h>
#include <stdint.h>
#include <time.h>
#include <sys/time.h>
#include <gpiod.h>
#include <sched.h>


#define IR_LIMITS  64  // bytes buffer = IR_LIMITS x8 bits
#define AEHA 21
#define NEC 22
#define SONY 23

unsigned char countHigh();
unsigned char countLow();
void getIR2();
void getIR();

unsigned char buf[IR_LIMITS];	// bytes buffer
unsigned char mode;	// 1:NEC, 2:AEHA, 3:SONY
unsigned int bits;	// 32768 bits capable


// irin  = GPIO4
// irout = GPIO13
#define IRIN  4
#define IROUT 13
struct gpiod_line *led;
struct gpiod_line *sw;


// delay nanosec with busy wait
// accuracy +100ns order on RPI5
// accuracy +300ns order on RPI4
// accuracy +1000ns order on RPI1
// more than 1000ns
// less than 1000000000ns
void delay(uint32_t t){

	struct timespec t0,t1;

	clock_gettime(CLOCK_MONOTONIC, &t0);
	do{
		clock_gettime(CLOCK_MONOTONIC, &t1);
	}while((t1.tv_sec - t0.tv_sec) *1000000000 + t1.tv_nsec - t0.tv_nsec < t);

}

unsigned char countHigh(){

	unsigned char i=0;

//	while( gpiod_line_get_value(sw)==1 );	// wait

	while( gpiod_line_get_value(sw)==0 ){
		++i;
		delay(26000);
		delay(26000);
		if(i==0) return 0;	// timeout
	}
	// NEC:i=19*8=152, i*2*26.5us=8056us
	// AEHA:i=19*4=76, i*2*26.5us=4028us
	// 1T:i=19*1=19

	return i;

}


unsigned char countLow(){

	unsigned char i=0;

//	while( gpiod_line_get_value(sw)==0 );	// wait

	while( gpiod_line_get_value(sw)==1 ){
		++i;
		delay(26000);
		if(i==0) return 0;	// timeout
	}
	// NEC:i=19*8=152, i*26.5us=4028us
	// AEHA:i=19*4=76, i*26.5us=2014us
	// 1T:i=19*1=19
	// 3T:i=19*3=57

	return i;

}


void getIR2(){

	unsigned char i;
	unsigned short j;   // capable 32768 bits = 4096 bytes
	unsigned char k;

	bits=0;
	for(j=0;j<IR_LIMITS;j++){  // buffer bytes LIMITS
		for(i=0;i<8;i++){   // 8 bits
			k = countHigh()*2;
			if(mode==3){
				buf[j]>>=1;
				// Threschold = 35, 23 = 1T, 46 = 2T; for SONY
				buf[j]+=((k>30) ? 0x80: 0);
				++bits;
			}
			k = countLow();
			if(k==0){
				buf[j]>>=(8-i);
				return;
			}
			if(mode!=3){
				buf[j]>>=1;
				// Threschold = 38, 19 = 1T, 57 = 3T; for NEC
				// Threschold = 30, 15 = 1T, 45 = 3T; for AEHA
				buf[j]+=((k>30) ? 0x80: 0);
				++bits;
			}
		}
	}

}


void display(unsigned char s)
{
        int i,n;
        char name[][5]={"none","NEC ","AEHA","SONY"};

        if(s==2)
        {
                printf("%s",name[mode]);    // 1:NEC, 2:AEHA, 3:SONY
                printf("%02X",bits);    // bits

                // no scroll
                n=(bits+7)/8;    // bits to n bytes
                for(i=0;i<n;i++){
                        printf("%02X",buf[i]);  // data
                }

        }else{

                printf("%s",name[mode]);    // 1:NEC, 2:AEHA, 3:SONY
                printf(" %d",bits);    // bits

                // scroll
                n=(bits+7)/8;    // bits to n bytes
                for(i=0;i<n;i++){
                        printf(" %d",buf[i]);  // data
                }
                printf("\n");
        }
}


void getIR(){

	unsigned char i;
	
	i = countHigh();    // Start
	mode=0;

//	printf("%d\n",i);
	if(40<i){
		if(i<51){
			mode=3; // SONY, 46
		}else{
			if(100<i){
				mode=1; // NEC, 173
			}else{
				mode=2; // AEHA, 54-77
			}
		}
		i = countLow();
		getIR2();
	}

}


int main(int argc, char *argv[])
{

	int minPriority,maxPriority;
	struct sched_param prio;
	unsigned char i,n;

	// to set high priority
	minPriority=sched_get_priority_min(SCHED_FIFO);
	maxPriority=sched_get_priority_max(SCHED_FIFO);
	prio.sched_priority=maxPriority;
	// printf("%d, %d\n",minPriority,maxPriority);
	if(sched_setscheduler(0,SCHED_FIFO,&prio) < 0){
		printf("error try sudo\n");
		return -1;
	}

        struct gpiod_chip *chip;
	// try gpiochip4 for RPI5
	chip = gpiod_chip_open_by_number(4);
 	if ( chip == NULL ) {
		// try gpiochip0
		chip = gpiod_chip_open_by_number(0);
	}
	led = gpiod_chip_get_line(chip, IROUT);
	sw = gpiod_chip_get_line(chip, IRIN);

	gpiod_line_request_output(led, "CONSUMER", 0);
	gpiod_line_set_value(led, 0);
	gpiod_line_request_falling_edge_events(sw, "CONSUMER");

	// to detect falling edge
	gpiod_line_event_wait(sw, NULL);
	getIR();
	display(argc);

	gpiod_chip_close(chip);
	return 0;

}


